/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import React, { useCallback, useEffect, useRef } from 'react';
import TWEEN, { Tween } from '@tweenjs/tween.js';
import { colors } from 'pc-nrfconnect-shared';
import { Mesh, MeshStandardMaterial, Vector3 } from 'three';
import shallow from 'zustand/shallow';

import { deviceStore, DirData, ReadingState } from '../stores/deviceStore';
import { settingsStore } from '../stores/settingsStore';
import { getEulerToGl } from '../util/calculations';
import { defaultForward } from './values';

enum CornerState {
    INAVTIVE,
    MEASURING,
    COMPLETED,
}

interface Corner {
    position: Vector3;
    devices: {
        index: number;
        directionBuffer: Vector3[];
    }[];
    state: CornerState;
}

const Calibrator: React.FC = () => {
    const coverageAngle = settingsStore(state => state.coverageAngle);
    const [
        readingState,
        devices,
        room,
        setReadingState,
        setCalibrationVectors,
    ] = deviceStore(
        state => [
            state.readingState,
            state.devices,
            state.roomSize,
            state.setReadingState,
            state.setCalibrationVectors,
        ],
        shallow
    );
    const corners = useRef<Corner[]>([]);
    const posRef = useRef<Mesh>();
    const matRef = useRef<MeshStandardMaterial>();
    const tweenRef = useRef<Tween<{ value: number }>>(
        new TWEEN.Tween({ value: 0 })
            .to({ value: 1 }, 500)
            .easing(TWEEN.Easing.Quartic.Out)
            .repeat(Infinity)
            .yoyo(true)
            .onUpdate((opacity: { value: number }) => {
                if (matRef.current) matRef.current.opacity = opacity.value;
            })
    );

    const findOffset = useCallback(
        (deviceIndex: number) => {
            const offset = [];
            for (let i = 0; i < corners.current.length; i += 1) {
                const device = corners.current[i].devices.find(
                    e => e.index === deviceIndex
                );
                if (device) {
                    const vector = new Vector3();
                    device.directionBuffer.forEach(e => vector.add(e));

                    vector
                        .divideScalar(device.directionBuffer.length)
                        .normalize();

                    offset.push({
                        measured: vector,
                        offset: corners.current[i].position
                            .clone()
                            .sub(devices[deviceIndex].position)
                            .normalize()
                            .sub(vector),
                    });
                } else {
                    // Add zero offset for undetected corners
                    offset.push({
                        measured: corners.current[i].position
                            .clone()
                            .sub(devices[deviceIndex].position)
                            .normalize(),
                        offset: new Vector3(0, 0, 0),
                    });
                }
            }
            return offset;
        },
        [corners, devices]
    );

    useEffect(() => {
        corners.current = [];
        corners.current.push({
            position: new Vector3(0, 0, 0),
            devices: [],
            state: CornerState.INAVTIVE,
        });
        corners.current.push({
            position: new Vector3(room.x, 0, 0),
            devices: [],
            state: CornerState.INAVTIVE,
        });
        corners.current.push({
            position: new Vector3(0, 0, room.z),
            devices: [],
            state: CornerState.INAVTIVE,
        });
        corners.current.push({
            position: new Vector3(room.x, 0, room.z),
            devices: [],
            state: CornerState.INAVTIVE,
        });
    }, [room]);

    useEffect(() => {
        switch (readingState) {
            case ReadingState.CALIBRATION_INITIALIZE:
                for (let i = 0; i < devices.length; i += 1) {
                    if (!devices[i].isDummy) {
                        const glDeviceForward = getEulerToGl(
                            devices[i].rotation
                        )(defaultForward);
                        for (let j = 0; j < corners.current.length; j += 1) {
                            const angle =
                                (glDeviceForward.angleTo(
                                    corners.current[j].position
                                        .clone()
                                        .sub(devices[i].position)
                                        .normalize()
                                ) *
                                    180) /
                                Math.PI;
                            if (angle <= coverageAngle) {
                                corners.current[j].devices.push({
                                    index: i,
                                    directionBuffer: [],
                                });
                            }
                        }
                    }
                }

                setReadingState(ReadingState.CALIBRATION_WAITING);
                break;
            case ReadingState.CALIBRATION_WAITING:
                for (let i = 0; i < corners.current.length; i += 1) {
                    if (
                        corners.current[i].devices.length > 0 &&
                        corners.current[i].state !== CornerState.COMPLETED
                    ) {
                        corners.current[i].state = CornerState.MEASURING;
                        posRef.current?.position.copy(
                            corners.current[i].position
                        );
                        tweenRef.current?.start();
                        return;
                    }
                }
                setReadingState(ReadingState.OFF);
                break;
            case ReadingState.CALIBRATION_RECORD:
                for (let i = 0; i < corners.current.length; i += 1) {
                    if (corners.current[i].state === CornerState.MEASURING) {
                        const unsubscribe = deviceStore.subscribe(
                            (dirData: DirData) => {
                                corners.current[i].devices
                                    .find(e => e.index === dirData.lastUpdate)
                                    ?.directionBuffer.push(
                                        dirData.deviceDirs[dirData.lastUpdate]
                                            .uncalibrated
                                    );
                            },
                            state => state.dirData
                        );
                        setTimeout(() => {
                            unsubscribe();
                            if (
                                corners.current[i].state ===
                                    CornerState.MEASURING &&
                                readingState === ReadingState.CALIBRATION_RECORD
                            ) {
                                corners.current[i].state =
                                    CornerState.COMPLETED;
                                for (
                                    let j = 0;
                                    j < corners.current.length;
                                    j += 1
                                ) {
                                    if (
                                        corners.current[j].state !==
                                            CornerState.COMPLETED &&
                                        corners.current[j].devices.length > 0
                                    ) {
                                        setReadingState(
                                            ReadingState.CALIBRATION_WAITING
                                        );
                                        return;
                                    }
                                }

                                for (let j = 0; j < devices.length; j += 1) {
                                    setCalibrationVectors(j, findOffset(j));
                                }
                            }
                            setReadingState(ReadingState.OFF);
                        }, 1000);
                    }
                }
                break;
            default:
                for (let i = 0; i < corners.current.length; i += 1) {
                    corners.current[i].devices = [];
                    corners.current[i].state = CornerState.INAVTIVE;
                }
                tweenRef.current?.stop();
                if (matRef.current) matRef.current.opacity = 0;
        }
    }, [
        readingState,
        devices,
        setReadingState,
        coverageAngle,
        tweenRef,
        findOffset,
        setCalibrationVectors,
    ]);

    return (
        <mesh ref={posRef}>
            <sphereBufferGeometry attach="geometry" args={[0.2, 32, 32]} />
            <meshStandardMaterial
                attach="material"
                color={colors.nordicBlue}
                ref={matRef}
                transparent
                opacity={0}
            />
        </mesh>
    );
};

export default Calibrator;
