/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import React, { useCallback, useEffect, useMemo, useRef } from 'react';
import TWEEN from '@tweenjs/tween.js';
import { colors } from 'pc-nrfconnect-shared';
import SerialPort from 'serialport';
import { ArrowHelper, Vector3 } from 'three';
import shallow from 'zustand/shallow';

import { DeviceDir, deviceStore } from '../../stores/deviceStore';
import { CalibrationVector } from '../../stores/deviceTypes';
import { settingsStore } from '../../stores/settingsStore';
import {
    applyCalibration,
    getEulerToGl,
    getEulerToLoc,
} from '../../util/calculations';
import { parseFrame } from '../../util/protocol';
import { defaultForward } from '../values';

const origin = new Vector3(0, 0, 0).normalize();

interface Props {
    index: number;
    position: Vector3;
    rotation: Vector3;
    // Won't be undefined as it is guaranteed through the deviceStore
    parser: SerialPort.parsers.Readline | undefined;
    calibrationVectors: CalibrationVector[];
}

const DeviceDirLine: React.FC<Props> = ({
    index,
    position,
    rotation,
    parser,
    calibrationVectors,
}) => {
    const [room, setDeviceDir] = deviceStore(
        state => [state.roomSize, state.setDeviceDir],
        shallow
    );

    const [
        showDirectionVectors,
        showRawVectors,
        useCalibrationVectors,
    ] = settingsStore(
        state => [
            state.showDirectionVectors,
            state.showRawVectors,
            state.useCalibrationVectors,
        ],
        shallow
    );

    const dirRef = useRef<ArrowHelper>();
    const secondaryDirRef = useRef<ArrowHelper>();
    // uncalibrated = local and calibrated = global for less vector rotation
    const lastDir = useRef<DeviceDir>({
        uncalibrated: new Vector3(),
        calibrated: new Vector3(),
    });
    const [eulerToGl, eulerToLoc] = useMemo(
        () => [getEulerToGl(rotation), getEulerToLoc(rotation)],
        [rotation]
    );

    const getLength = useCallback(
        (vector: Vector3) => {
            let xScalar = 0;
            let yScalar = 0;
            let zScalar = 0;
            if (vector.x > 0) xScalar = (room.x - position.x) / vector.x;
            else xScalar = Math.abs(position.x / vector.x);
            if (vector.y > 0) yScalar = (room.y - position.y) / vector.y;
            else yScalar = Math.abs(position.y / vector.y);
            if (vector.z > 0) zScalar = (room.z - position.z) / vector.z;
            else zScalar = Math.abs(position.z / vector.z);

            const minScalar = Math.min(xScalar, yScalar, zScalar);

            return Math.sqrt(
                Math.abs(vector.x * minScalar) ** 2 +
                    Math.abs(vector.y * minScalar) ** 2 +
                    Math.abs(vector.z * minScalar) ** 2
            );
        },
        [room, position]
    );

    useEffect(() => {
        // Guarantee there are no other listeners before assigning
        parser?.removeAllListeners();
        const parseMessage = parseFrame();
        parser?.on('data', m => {
            const dirLoc = parseMessage(m);
            if (dirLoc) {
                const calDirGl = applyCalibration(
                    eulerToGl(dirLoc),
                    calibrationVectors
                );

                new TWEEN.Tween({ ...lastDir.current })
                    .to({ uncalibrated: dirLoc, calibrated: calDirGl }, 150)
                    .easing(TWEEN.Easing.Quadratic.Out)
                    .onUpdate(deviceDir => {
                        lastDir.current = deviceDir;
                        const uncalDirGl = eulerToGl(deviceDir.uncalibrated);
                        setDeviceDir(index, {
                            uncalibrated: uncalDirGl,
                            calibrated: deviceDir.calibrated,
                        });

                        if (showDirectionVectors) {
                            dirRef.current?.setDirection(
                                useCalibrationVectors
                                    ? eulerToLoc(deviceDir.calibrated)
                                    : deviceDir.uncalibrated
                            );
                            dirRef.current?.setLength(
                                getLength(
                                    useCalibrationVectors
                                        ? deviceDir.calibrated
                                        : uncalDirGl
                                ),
                                0.03,
                                0.05
                            );

                            if (useCalibrationVectors && showRawVectors) {
                                secondaryDirRef.current?.setDirection(
                                    deviceDir.uncalibrated
                                );
                                secondaryDirRef.current?.setLength(
                                    getLength(uncalDirGl),
                                    0.03,
                                    0.05
                                );
                            }
                        }
                    })
                    .start();
            }
        });
        // Can't run removeAllListeners in cleanup because it runs after ReadingsPane setup where a listener is setup
    }, [
        index,
        setDeviceDir,
        getLength,
        showDirectionVectors,
        calibrationVectors,
        useCalibrationVectors,
        showRawVectors,
        eulerToGl,
        eulerToLoc,
        parser,
    ]);

    if (!showDirectionVectors) {
        return <></>;
    }

    return (
        <group>
            <arrowHelper
                ref={dirRef}
                args={[
                    defaultForward.clone(),
                    origin,
                    getLength(eulerToGl(defaultForward)),
                    colors.purple,
                    0.03,
                    0.05,
                ]}
            />
            {useCalibrationVectors && showRawVectors && (
                <arrowHelper
                    ref={secondaryDirRef}
                    args={[
                        defaultForward.clone(),
                        origin,
                        getLength(eulerToGl(defaultForward)),
                        colors.orange,
                        0.03,
                        0.05,
                    ]}
                />
            )}
        </group>
    );
};

export default DeviceDirLine;
