/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import React, { useEffect, useMemo } from 'react';
import { colors, getAppDir } from 'pc-nrfconnect-shared';
import { MeshBasicMaterial, TextureLoader, Vector3 } from 'three';
import shallow from 'zustand/shallow';

import { displayedDeviceName } from '../../SidePanel/Device/DeviceHeader/DeviceInfo';
import { deviceStore, ReadingState } from '../../stores/deviceStore';
import { Device } from '../../stores/deviceTypes';
import { settingsStore } from '../../stores/settingsStore';
import { getEulerToGl, vectorRotationToEuler } from '../../util/calculations';
import { defaultForward, DEVICE_SIDE, DEVICE_THICKNESS } from '../values';
import DeviceCone from './DeviceCone';
import DeviceDirLine from './DeviceDirLine';

interface Props {
    device: Device;
    index: number;
}

const origin = new Vector3(0, 0, 0).normalize();
const length = DEVICE_SIDE * 1.2;
const headLength = 0.2 * length;
const headWidth = 0.8 * headLength;

const textureLoader = new TextureLoader();
const front = textureLoader.load(`${getAppDir()}/resources/Front.png`);
const back = textureLoader.load(`${getAppDir()}/resources/Back.png`);

const deviceMaterial = [
    new MeshBasicMaterial({ color: colors.nordicBlue }),
    new MeshBasicMaterial({ color: colors.nordicBlue }),
    new MeshBasicMaterial({ color: colors.nordicBlue }),
    new MeshBasicMaterial({ color: colors.nordicBlue }),
    new MeshBasicMaterial({ map: back }),
    new MeshBasicMaterial({ map: front }),
];

const DeviceGraphic: React.FC<Props> = ({ device, index }) => {
    const [showCoverage, showLabels] = settingsStore(
        state => [state.showCoverage, state.showLabels],
        shallow
    );

    const [readingState, setDeviceDir] = deviceStore(
        state => [state.readingState, state.setDeviceDir],
        shallow
    );

    const eulerToGl = getEulerToGl(device.rotation);
    const canvas = document.createElement('canvas');
    const centerX = canvas.width / 2;
    const centerY = canvas.height / 2;
    const context = canvas.getContext('2d');
    if (context && showLabels) {
        context.textBaseline = 'bottom';
        context.font = `bold 35px helvetica`;
        context.lineWidth = 4;
        context.fillStyle = 'black';
        context.textAlign = 'center';
        context.fillText(displayedDeviceName(device), centerX, centerY);
    }

    const rotation = useMemo(() => vectorRotationToEuler(device.rotation), [
        device.rotation,
    ]);

    // Set default direction in deviceStore
    useEffect(() => {
        const glDir = eulerToGl(defaultForward);
        setDeviceDir(index, { uncalibrated: glDir, calibrated: glDir });
    }, [index, setDeviceDir, eulerToGl]);

    return (
        <group position={device.position}>
            {showLabels && (
                <group position={[0, DEVICE_SIDE / 2, 0]}>
                    <sprite scale={[3, 1.5, 3]}>
                        <spriteMaterial attach="material">
                            <canvasTexture attach="map" image={canvas} />
                        </spriteMaterial>
                    </sprite>
                </group>
            )}
            <group rotation={rotation}>
                <mesh material={deviceMaterial}>
                    <boxBufferGeometry
                        attach="geometry"
                        args={[DEVICE_SIDE, DEVICE_SIDE, DEVICE_THICKNESS]}
                    />
                </mesh>
                {(readingState !== ReadingState.OFF && (
                    <DeviceDirLine
                        index={index}
                        position={device.position}
                        rotation={device.rotation}
                        parser={device.parser}
                        calibrationVectors={device.calibrationVectors}
                    />
                )) ||
                    (!showCoverage && (
                        <arrowHelper
                            args={[
                                defaultForward.clone(),
                                origin,
                                length,
                                colors.red,
                                headLength,
                                headWidth,
                            ]}
                        />
                    )) || <DeviceCone />}
            </group>
        </group>
    );
};

export default DeviceGraphic;
