/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import { useEffect } from 'react';
import { Group, Vector2 } from 'three';

// A lot of this was taken from the OrbitControls module included
// in three.js. It's been adapted to rotate a mesh rather than
// the camera.

export interface UseMouseOptions {
    enabled: boolean;
    enableRotateX: boolean;
    enableRotateY: boolean;
}

function useMouse(
    el: HTMLElement,
    group: Group | undefined,
    options: UseMouseOptions
) {
    useEffect(() => {
        let isMouseDown = false;

        const rotationSpeed = 0.5;
        const rotateStart = new Vector2();
        const rotateEnd = new Vector2();
        const rotateDelta = new Vector2();

        function rotateX(angle: number) {
            if (!options.enableRotateX) return;
            if (!group) return;
            group.rotation.y += angle;
        }

        function rotateY(angle: number) {
            if (!options.enableRotateY) return;
            if (!group) return;

            const oldRotation = group.rotation.x;
            const newRotation = group.rotation.x + angle;

            // Lock the movement to the two half-positions, to stop
            // the user viewing "underneath" the dome.
            const movingDown = oldRotation > newRotation;
            if (movingDown && newRotation > 0) {
                group.rotation.x = newRotation;
            } else if (!movingDown && newRotation < Math.PI) {
                group.rotation.x = newRotation;
            }
        }

        function mouseUpHandler() {
            isMouseDown = false;
        }

        function mouseLeaveHandler() {
            isMouseDown = false;
        }

        function mouseDownHandler(e: MouseEvent) {
            isMouseDown = true;
            rotateStart.set(e.clientX, e.clientY);
        }

        function mouseMoveHandler(e: MouseEvent) {
            if (!isMouseDown) return;

            rotateEnd.set(e.clientX, e.clientY);
            rotateDelta
                .subVectors(rotateEnd, rotateStart)
                .multiplyScalar(rotationSpeed);

            rotateX((2 * Math.PI * rotateDelta.x) / el.clientHeight);
            rotateY((2 * Math.PI * rotateDelta.y) / el.clientHeight);

            rotateStart.copy(rotateEnd);
        }

        function registerMouseListeners() {
            el.addEventListener('mouseup', mouseUpHandler);
            el.addEventListener('mousedown', mouseDownHandler);
            el.addEventListener('mousemove', mouseMoveHandler);
            el.addEventListener('mouseleave', mouseLeaveHandler);
        }

        function removeMouseListeners() {
            el.removeEventListener('mouseup', mouseUpHandler);
            el.removeEventListener('mousedown', mouseDownHandler);
            el.removeEventListener('mousemove', mouseMoveHandler);
            el.removeEventListener('mouseleave', mouseLeaveHandler);
        }

        if (options.enabled) {
            registerMouseListeners();
        }

        return removeMouseListeners;
    }, [el, group, options]);
}

export default useMouse;
