/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import { useEffect } from 'react';
import { useThree } from 'react-three-fiber';
import { Group } from 'three';

import { CAMERA_Z_OFFSET } from '../values';
import useKeyPress from './useKeyPress';
import useMouse, { UseMouseOptions } from './useMouse';

// Show disconnect and time since last packet in the ui, and go red if it's been more than two seconds
export interface KeyboardOptions {
    enabled: boolean;
    enableRotateX: boolean;
    enableRotateY: boolean;
}

interface UseMovementOptions {
    mouse: UseMouseOptions;
    keyboard: KeyboardOptions;
    horizontalSpeed: number;
    verticalSpeed: number;
    zoomSpeed: number;
    maxZoom: number;
}

function useMovement(group: Group | undefined, options: UseMovementOptions) {
    const { camera, gl } = useThree();
    const el = gl.domElement;

    useMouse(el, group, options.mouse);

    const wPress = useKeyPress('w');
    const sPress = useKeyPress('s');
    const aPress = useKeyPress('a');
    const dPress = useKeyPress('d');

    const upPress = useKeyPress('ArrowUp');
    const downPress = useKeyPress('ArrowDown');
    const leftPress = useKeyPress('ArrowLeft');
    const rightPress = useKeyPress('ArrowRight');

    const zPress = useKeyPress('z');
    const xPress = useKeyPress('x');

    const MIN_ZOOM = CAMERA_Z_OFFSET;
    const MAX_ZOOM = options.maxZoom;

    function keyboardEvents() {
        const { enabled, enableRotateX, enableRotateY } = options.keyboard;

        if (!group) return;
        if (!enabled) return;

        if (enableRotateY) {
            if (
                (wPress.current || upPress.current) &&
                group.rotation.x < Math.PI
            ) {
                group.rotation.x += options.verticalSpeed;
            }
            if (
                (sPress.current || downPress.current) &&
                group.rotation.x > 0.0
            ) {
                group.rotation.x -= options.verticalSpeed;
            }
        }

        if (enableRotateX) {
            if (aPress.current || leftPress.current) {
                group.rotation.y += options.horizontalSpeed;
            }
            if (dPress.current || rightPress.current) {
                group.rotation.y -= options.horizontalSpeed;
            }
        }

        if (camera.position.z > MAX_ZOOM) {
            camera.position.z = MAX_ZOOM;
        }

        if (
            zPress.current &&
            camera.position.z - options.zoomSpeed >= MIN_ZOOM
        ) {
            camera.position.z -= options.zoomSpeed;
        } else if (xPress.current && camera.position.z < MAX_ZOOM) {
            camera.position.z += options.zoomSpeed;
        }
    }

    useEffect(() => {
        el.focus();
    }, [el]);

    // Called during the main render loop.
    return function handleMovement() {
        if (options?.keyboard?.enabled) {
            keyboardEvents();
        }
    };
}

export default useMovement;
