/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import React, { useEffect, useMemo, useRef } from 'react';
import { useFrame } from 'react-three-fiber';
import { colors } from 'pc-nrfconnect-shared';
import { BufferGeometry, Mesh, Vector3 } from 'three';
import shallow from 'zustand/shallow';

import { deviceStore, DirData } from '../stores/deviceStore';
import { settingsStore } from '../stores/settingsStore';
import { getPosition } from '../util/calculations';

const nullVector = new Vector3(0, 0, 0);

function XPoints() {
    const points = [new Vector3(), new Vector3()];
    return (pos: Vector3, room: Vector3) => {
        points[0].set(0, pos.y, pos.z);
        points[1].set(room.x, pos.y, pos.z);
        return points;
    };
}

function ZPoints() {
    const points = [new Vector3(), new Vector3()];
    return (pos: Vector3, room: Vector3) => {
        points[0].set(pos.x, pos.y, 0);
        points[1].set(pos.x, pos.y, room.z);
        return points;
    };
}

const PositionIndicator: React.FC = () => {
    const [devices, room] = deviceStore(
        state => [state.devices, state.roomSize],
        shallow
    );
    const [useCalibrationVectors, showRawVectors] = settingsStore(
        state => [state.useCalibrationVectors, state.showRawVectors],
        shallow
    );

    const deviceDirsRef = useRef<DirData>(deviceStore.getState().dirData);
    const getXPoints = XPoints();
    const getZPoints = ZPoints();
    const posRef = useRef<Mesh>();
    const shadowRef = useRef<Mesh>();
    const lineX = useRef<BufferGeometry>();
    const lineZ = useRef<BufferGeometry>();
    const secondaryPosRef = useRef<Mesh>();
    const secondaryShadowRef = useRef<Mesh>();
    const secondaryLineX = useRef<BufferGeometry>();
    const secondaryLineZ = useRef<BufferGeometry>();

    const positions = useMemo(() => devices.map(e => e.position.clone()), [
        devices,
    ]);

    useEffect(
        () =>
            deviceStore.subscribe(
                (dirData: DirData) => {
                    deviceDirsRef.current = dirData;
                },
                state => state.dirData
            ),
        [positions, room]
    );

    function updateMainPos(pos: Vector3) {
        posRef.current?.position.copy(pos);
        shadowRef.current?.position.setX(pos.x).setZ(pos.z);
        lineX.current?.setFromPoints(getXPoints(pos, room));
        lineZ.current?.setFromPoints(getZPoints(pos, room));
    }

    function updateSecondaryPos(pos: Vector3) {
        secondaryPosRef.current?.position.copy(pos);
        secondaryShadowRef.current?.position.setX(pos.x).setZ(pos.z);
        secondaryLineX.current?.setFromPoints(getXPoints(pos, room));
        secondaryLineZ.current?.setFromPoints(getZPoints(pos, room));
    }

    useFrame(() => {
        if (deviceDirsRef.current.deviceDirs.length >= 2) {
            if (useCalibrationVectors) {
                updateMainPos(
                    getPosition(
                        positions,
                        deviceDirsRef.current.deviceDirs.map(e => e.calibrated)
                    ).clamp(nullVector, room)
                );

                if (showRawVectors) {
                    updateSecondaryPos(
                        getPosition(
                            positions,
                            deviceDirsRef.current.deviceDirs.map(
                                e => e.uncalibrated
                            )
                        ).clamp(nullVector, room)
                    );
                }
            } else {
                updateMainPos(
                    getPosition(
                        positions,
                        deviceDirsRef.current.deviceDirs.map(
                            e => e.uncalibrated
                        )
                    ).clamp(nullVector, room)
                );
            }
        }
    });

    return (
        <group>
            <mesh ref={posRef}>
                <sphereBufferGeometry attach="geometry" args={[0.1, 32, 32]} />
                <meshStandardMaterial attach="material" color={colors.red} />
            </mesh>
            <lineSegments>
                <bufferGeometry attach="geometry" ref={lineX} />
                <lineBasicMaterial
                    attach="material"
                    color={colors.black}
                    opacity={0.4}
                />
            </lineSegments>
            <lineSegments>
                <bufferGeometry attach="geometry" ref={lineZ} />
                <lineBasicMaterial
                    attach="material"
                    color={colors.black}
                    opacity={0.4}
                />
            </lineSegments>
            <mesh ref={shadowRef} rotation={[-Math.PI / 2, 0, 0]}>
                <circleBufferGeometry attach="geometry" args={[0.1, 32]} />
                <meshStandardMaterial
                    attach="material"
                    color={colors.black}
                    opacity={0.4}
                />
            </mesh>
            {useCalibrationVectors && showRawVectors && (
                <group>
                    <mesh ref={secondaryPosRef}>
                        <sphereBufferGeometry
                            attach="geometry"
                            args={[0.1, 32, 32]}
                        />
                        <meshStandardMaterial
                            attach="material"
                            color={colors.orange}
                        />
                    </mesh>
                    <lineSegments>
                        <bufferGeometry
                            attach="geometry"
                            ref={secondaryLineX}
                        />
                        <lineBasicMaterial
                            attach="material"
                            color={colors.black}
                            opacity={0.3}
                        />
                    </lineSegments>
                    <lineSegments>
                        <bufferGeometry
                            attach="geometry"
                            ref={secondaryLineZ}
                        />
                        <lineBasicMaterial
                            attach="material"
                            color={colors.black}
                            opacity={0.3}
                        />
                    </lineSegments>
                    <mesh
                        ref={secondaryShadowRef}
                        rotation={[-Math.PI / 2, 0, 0]}
                    >
                        <circleBufferGeometry
                            attach="geometry"
                            args={[0.1, 32]}
                        />
                        <meshStandardMaterial
                            attach="material"
                            color={colors.black}
                            opacity={0.3}
                        />
                    </mesh>
                </group>
            )}
        </group>
    );
};

export default PositionIndicator;
