/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import React, { useCallback, useEffect, useMemo, useRef } from 'react';
import { useFrame, useThree } from 'react-three-fiber';
import { colors } from 'pc-nrfconnect-shared';
import { Group } from 'three';
import shallow from 'zustand/shallow';

import { deviceStore, ReadingState } from '../stores/deviceStore';
import Calibrator from './Calibrator';
import DeviceCollection from './Device/DeviceCollection';
import { Event, graphicEmitter, Perspective } from './Perspective/events';
import useMovement from './Perspective/useMovement';
import PositionIndicator from './PositionIndicator';
import Room from './Room';
import { CAMERA_Z_OFFSET, ROTATION_FACTOR } from './values';

const perspectivePresets = {
    [Perspective.Side]: {
        x: 0,
        y: ROTATION_FACTOR,
        z: 0,
    },
    [Perspective.Default]: {
        x: 0.35,
        y: ROTATION_FACTOR,
        z: 0,
    },
    [Perspective.Top]: {
        x: 1.5,
        y: ROTATION_FACTOR,
        z: 0,
    },
};

const PositionViewer: React.FC = () => {
    const { camera } = useThree();
    const [roomSize, readingState, setReadingState] = deviceStore(
        state => [state.roomSize, state.readingState, state.setReadingState],
        shallow
    );

    // Explicitly set to OFF to avoid bug when f.ex. in calibration_waiting and switching to Readings pane and back
    useEffect(() => () => setReadingState(ReadingState.OFF), [setReadingState]);

    const gridSize = useMemo(() => Math.max(roomSize.x, roomSize.z) + 4, [
        roomSize,
    ]);
    const gridColour = colors.gray200;

    const groupRef = useRef<Group>();
    const group = groupRef.current;

    const maxZoom = useMemo(
        () => Math.max(...roomSize.toArray()) * 0.7 + CAMERA_Z_OFFSET,
        [roomSize]
    );

    const handleMovement = useMovement(group, {
        keyboard: {
            enabled: true,
            enableRotateX: true,
            enableRotateY: true,
        },
        mouse: {
            enabled: true,
            enableRotateX: true,
            enableRotateY: true,
        },
        horizontalSpeed: 0.04,
        verticalSpeed: 0.0225,
        zoomSpeed: 0.2,
        maxZoom,
    });

    const setPerspective = useCallback(
        (perspective: Perspective) => {
            camera.position.set(0, 0, maxZoom);
            camera.updateProjectionMatrix();
            group?.rotation.set(
                perspectivePresets[perspective].x,
                perspectivePresets[perspective].y,
                perspectivePresets[perspective].z
            );
        },
        [camera, group?.rotation?.set, maxZoom]
    );

    const keyHandlers = useCallback(
        ({ key }: KeyboardEvent) => {
            // Also supports numpad keys.
            if (key === '1' || key === '97') {
                setPerspective(Perspective.Default);
            } else if (key === '2' || key === '98') {
                setPerspective(Perspective.Top);
            } else if (key === '3' || key === '99') {
                setPerspective(Perspective.Side);
            }
        },
        [setPerspective]
    );

    const addPerspectiveListeners = useCallback(() => {
        window.addEventListener('keyup', keyHandlers);
        graphicEmitter.on(Event.SET_PERSPECTIVE, setPerspective);
    }, [keyHandlers, setPerspective]);

    const removePerspectiveListeners = useCallback(() => {
        window.removeEventListener('keyup', keyHandlers);
        graphicEmitter.removeListener(Event.SET_PERSPECTIVE, setPerspective);
    }, [keyHandlers, setPerspective]);

    useEffect(() => {
        setPerspective(Perspective.Default);
        addPerspectiveListeners();
        return () => removePerspectiveListeners();
    }, [addPerspectiveListeners, removePerspectiveListeners, setPerspective]);

    useFrame(() => {
        handleMovement();
    });

    return (
        <group ref={groupRef}>
            <gridHelper args={[gridSize, gridSize, gridColour, gridColour]} />
            <group position={[-(roomSize.x / 2), 0.01, -(roomSize.z / 2)]}>
                <Room />
                <DeviceCollection />
                {readingState !== ReadingState.OFF && <PositionIndicator />}
                <Calibrator />
            </group>
        </group>
    );
};

export default PositionViewer;
