/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import React, { useEffect, useRef, useState } from 'react';
import SerialPort from 'serialport';
import { Vector3 } from 'three';

import { defaultForward } from '../Position/values';
import { displayedDeviceName } from '../SidePanel/Device/DeviceHeader/DeviceInfo';
import { DeviceDir, deviceStore } from '../stores/deviceStore';
import {
    applyCalibration,
    getClosestPointsVectors,
    getEulerToGl,
} from '../util/calculations';
import { parseFrame } from '../util/protocol';
import OffsetChart from './OffsetChart';

const MAX_FRAMES = 35;

interface OffsetPair {
    index1: number;
    index2: number;
    uncal: number[];
    cal: number[];
}

function getDistance(vector1: Vector3, vector2: Vector3) {
    return Math.sqrt(
        Math.abs(vector1.x - vector2.x) ** 2 +
            Math.abs(vector1.y - vector2.y) ** 2 +
            Math.abs(vector1.z - vector2.z) ** 2
    );
}

const ChartCollection: React.FC = () => {
    const devices = deviceStore(state => state.devices);
    const runInterval = useRef<NodeJS.Timeout>();
    const devicePort = useRef<SerialPort[]>([]);
    const parser = useRef<SerialPort.parsers.Readline[]>([]);
    const deviceDir = useRef<DeviceDir[]>([]);
    const [pairs, setPairs] = useState<OffsetPair[]>([]);

    useEffect(() => {
        const temp = [];
        for (let i = 0; i < devices.length - 1; i += 1) {
            for (let j = i + 1; j < devices.length; j += 1) {
                temp.push({ index1: i, index2: j, uncal: [], cal: [] });
            }
        }
        setPairs(temp);
    }, [devices]);

    useEffect(() => {
        for (let i = 0; i < devices.length; i += 1) {
            const eulerToGl = getEulerToGl(devices[i].rotation);
            deviceDir.current.push({
                uncalibrated: new Vector3(),
                calibrated: new Vector3(),
            });
            if (!devices[i].isDummy) {
                const parseMessage = parseFrame();
                // Guarantee there are no other listeners before assigning
                devices[i].parser?.removeAllListeners();
                devices[i].parser?.on('data', m => {
                    const uncalDirLoc = parseMessage(m);
                    if (uncalDirLoc) {
                        const uncalDirGl = eulerToGl(uncalDirLoc);
                        deviceDir.current[i] = {
                            uncalibrated: uncalDirGl,
                            calibrated: applyCalibration(
                                uncalDirGl,
                                devices[i].calibrationVectors
                            ),
                        };
                    }
                });
            } else {
                const forward = eulerToGl(defaultForward);
                deviceDir.current[i] = {
                    uncalibrated: forward,
                    calibrated: forward,
                };
            }
        }

        return () => {
            setPairs([]);
            devices.forEach(e => {
                e.parser?.removeAllListeners();
            });
        };
    }, [devices, devicePort, parser]);

    useEffect(() => {
        runInterval.current = setInterval(() => {
            setPairs(
                pairs.map(e => {
                    if (e.uncal.length > MAX_FRAMES) {
                        e.uncal.shift();
                    }
                    if (e.cal.length > MAX_FRAMES) {
                        e.cal.shift();
                    }

                    const uncal = getClosestPointsVectors(
                        [
                            devices[e.index1].position,
                            devices[e.index2].position,
                        ],
                        [
                            deviceDir.current[e.index1].uncalibrated,
                            deviceDir.current[e.index2].uncalibrated,
                        ]
                    );
                    const cal = getClosestPointsVectors(
                        [
                            devices[e.index1].position,
                            devices[e.index2].position,
                        ],
                        [
                            deviceDir.current[e.index1].calibrated,
                            deviceDir.current[e.index2].calibrated,
                        ]
                    );

                    return {
                        ...e,
                        uncal: [...e.uncal, getDistance(uncal[0], uncal[1])],
                        cal: [...e.cal, getDistance(cal[0], cal[1])],
                    };
                })
            );
        }, 500);

        return () => {
            if (runInterval.current) clearInterval(runInterval.current);
        };
    }, [runInterval, pairs, devices]);

    return (
        <>
            {pairs.map(e => (
                // key: index + 1 to avoid duplicates with index 0
                <div
                    className="chart-wrapper"
                    key={(e.index1 + 1) * (e.index2 + 1)}
                >
                    <OffsetChart
                        name1={displayedDeviceName(devices[e.index1])}
                        name2={displayedDeviceName(devices[e.index2])}
                        uncal={e.uncal}
                        cal={e.cal}
                    />
                </div>
            ))}
        </>
    );
};

export default ChartCollection;
