/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import React from 'react';
import { colors, Group, NumberInlineInput, Toggle } from 'pc-nrfconnect-shared';
import shallow from 'zustand/shallow';

import { MAX_DETECTION_ANGLE } from '../Position/values';
import { deviceStore, ReadingState } from '../stores/deviceStore';
import { settingsStore } from '../stores/settingsStore';

const Settings: React.FC = () => {
    const [
        showAxes,
        showLabels,
        showCoverage,
        angle,
        showDirectionVectors,
        useCalibrationVectors,
        showRawVectors,
        toggleAxes,
        toggleLabels,
        toggleCoverage,
        setCoverageAngle,
        setUseMovement,
        toggleDirectionVectors,
        toggleCalibrationVectors,
        toggleRawVectors,
    ] = settingsStore(
        state => [
            state.showAxes,
            state.showLabels,
            state.showCoverage,
            state.coverageAngle,
            state.showDirectionVectors,
            state.useCalibrationVectors,
            state.showRawVectors,
            state.toggleAxes,
            state.toggleLabels,
            state.toggleCoverage,
            state.setCoverageAngle,
            state.setUseMovement,
            state.toggleDirectionVectors,
            state.toggleCalibrationVectors,
            state.toggleRawVectors,
        ],
        shallow
    );

    const readingState = deviceStore(state => state.readingState);

    return (
        <Group heading="Settings">
            <Toggle
                title="Turn on/off 3D axes"
                label="Show 3D axes"
                barColor={colors.gray700}
                // @ts-ignore Mismatch between Toggle component and index.d file. Fixed in later version of pc-nrfconnect-shared
                barColorToggled={colors.nordicBlue}
                variant="secondary"
                isToggled={showAxes}
                onToggle={toggleAxes}
            />
            <Toggle
                title="Turn on/off visibility of device labels"
                label="Show device labels"
                barColor={colors.gray700}
                // @ts-ignore Mismatch between Toggle component and index.d file. Fixed in later version of pc-nrfconnect-shared
                barColorToggled={colors.nordicBlue}
                variant="secondary"
                isToggled={showLabels}
                onToggle={toggleLabels}
            />
            <Toggle
                title="Turn on/off device coverage"
                label="Show device coverage"
                barColor={colors.gray700}
                // @ts-ignore Mismatch between Toggle component and index.d file. Fixed in later version of pc-nrfconnect-shared
                barColorToggled={colors.nordicBlue}
                variant="secondary"
                isToggled={showCoverage}
                onToggle={toggleCoverage}
            />
            <div className="inline-space-between">
                <p className="settings-label">Locator detection angle</p>
                <div
                    className="inline"
                    onFocus={() => setUseMovement(false)}
                    onBlur={() => setUseMovement(true)}
                >
                    <NumberInlineInput
                        value={angle}
                        range={{
                            min: 1,
                            max: MAX_DETECTION_ANGLE,
                            decimals: 0,
                        }}
                        disabled={readingState !== ReadingState.OFF}
                        onChange={val => setCoverageAngle(val)}
                    />
                    <p>°</p>
                </div>
            </div>
            <Toggle
                title="Turn on/off direction vectors during reading"
                label="Show direction vectors"
                barColor={colors.gray700}
                // @ts-ignore Mismatch between Toggle component and index.d file. Fixed in later version of pc-nrfconnect-shared
                barColorToggled={colors.nordicBlue}
                variant="secondary"
                isToggled={showDirectionVectors}
                onToggle={toggleDirectionVectors}
            />
            <Toggle
                title="Turn on/off using calibration vectors"
                label="Use calibration vectors"
                barColor={colors.gray700}
                // @ts-ignore Mismatch between Toggle component and index.d file. Fixed in later version of pc-nrfconnect-shared
                barColorToggled={colors.nordicBlue}
                variant="secondary"
                isToggled={useCalibrationVectors}
                onToggle={toggleCalibrationVectors}
            />
            {useCalibrationVectors && showDirectionVectors && (
                <Toggle
                    title="Turn on/off display of raw direction vectors"
                    label="Show raw vectors"
                    barColor={colors.gray700}
                    // @ts-ignore Mismatch between Toggle component and index.d file. Fixed in later version of pc-nrfconnect-shared
                    barColorToggled={colors.nordicBlue}
                    variant="secondary"
                    isToggled={showRawVectors}
                    onToggle={toggleRawVectors}
                />
            )}
        </Group>
    );
};

export default Settings;
