/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import { Dispatch } from 'react';
import produce, { Draft } from 'immer';
import { Device as BaseDevice } from 'pc-nrfconnect-shared';
import { Vector3 } from 'three';
import create, { State, StateCreator } from 'zustand';

import { deviceSetup } from '../DeviceSelector';
import { getDevicePort, Readline } from '../util/protocol';
import { CalibrationVector, Device } from './deviceTypes';
import {
    getPersistedDevice,
    getPersistedRoom,
    persistDevices,
    persistRoom,
} from './persistentStore';

const DeviceSetup = require('nrf-device-setup');

enum ActionType {
    DEVICE_DESELECTED = 'DEVICE_DESELECTED',
}

export interface DeselectDevice {
    type: ActionType.DEVICE_DESELECTED;
}

export enum ReadingState {
    OFF,
    READING,
    CALIBRATION_INITIALIZE,
    CALIBRATION_WAITING,
    CALIBRATION_RECORD,
}

export interface DeviceDir {
    uncalibrated: Vector3;
    calibrated: Vector3;
}

export interface DirData {
    lastUpdate: number;
    deviceDirs: DeviceDir[];
}

function defaultDevice(base: BaseDevice) {
    return {
        ...base,
        position: new Vector3(),
        rotation: new Vector3(),
        correctFirmware: false,
        isDummy: false,
        calibrationVectors: [],
        parser: undefined,
        port: undefined,
    };
}

type DeviceStore = {
    devices: Device[];
    readingState: ReadingState;
    dirData: DirData;
    roomSize: Vector3;
    setRoomSize: (size: Vector3) => void;
    setReadingState: (status: ReadingState) => void;
    addDevice: (
        device: BaseDevice,
        dispatch?: Dispatch<DeselectDevice>
    ) => void;
    removeDevice: (index: number) => void;
    setCalibrationVectors: (
        index: number,
        vectors: CalibrationVector[]
    ) => void;
    setCorrectFirmware: (index: number) => void;
    setDevicePosition: (index: number, position: Vector3) => void;
    setDeviceRotation: (index: number, rotation: Vector3) => void;
    setDeviceDir: (index: number, dir: DeviceDir) => void;
};

const immer = <T extends State>(
    config: StateCreator<T, (fn: (draft: Draft<T>) => void) => void>
): StateCreator<T> => (set, get, api) =>
    config(fn => set(produce(fn) as (state: T) => T), get, api);

const nullVector = new Vector3(0, 0, 0);

export const deviceStore = create<DeviceStore>(
    immer((set, get) => ({
        devices: [],
        readingState: ReadingState.OFF,
        dirData: { lastUpdate: -1, deviceDirs: [] },
        roomSize: new Vector3().copy(getPersistedRoom()),
        setReadingState: (status: ReadingState) =>
            set(state => {
                state.readingState = status;
            }),
        setRoomSize: size => {
            set(state => {
                state.roomSize = size;
            });

            // clamp devicepositions so they are always valid
            // clear calibration vectors since room has changed
            set(state =>
                state.devices.forEach(device => {
                    device.calibrationVectors = [];
                    device.position = device.position
                        .clone()
                        .clamp(nullVector, state.roomSize);
                })
            );
            persistRoom(get().roomSize);
        },
        addDevice: (device, dispatch) => {
            if (get().readingState >= ReadingState.CALIBRATION_INITIALIZE) {
                set(state => {
                    state.readingState = ReadingState.OFF;
                });
            }

            // Add dummy device
            if (!device.serialport) {
                set(state => {
                    state.devices.push({
                        ...defaultDevice(device),
                        traits: [...device.traits],
                        isDummy: true,
                        correctFirmware: true,
                    });
                    state.dirData.deviceDirs.push({
                        uncalibrated: new Vector3(),
                        calibrated: new Vector3(),
                    });
                });
            } else if (
                !get().devices.some(e => e.serialNumber === device.serialNumber)
            ) {
                const newDevice: Device = {
                    ...defaultDevice(device),
                    traits: [...device.traits],
                };

                const savedDevice = getPersistedDevice(device.serialNumber);
                if (savedDevice) {
                    // Have to clone all Vector3 objects to avoid a bug
                    newDevice.position
                        .copy(savedDevice.position)
                        .clamp(nullVector, get().roomSize);
                    newDevice.rotation.copy(savedDevice.rotation);
                    newDevice.calibrationVectors = savedDevice.calibrationVectors.map(
                        e => ({
                            measured: new Vector3().copy(e.measured),
                            offset: new Vector3().copy(e.offset),
                        })
                    );
                }

                // check if device has correct firmware but do not program
                const deviceSetupConfig = {
                    promiseConfirm: () =>
                        new Promise(resolve => {
                            set(state => {
                                state.devices.push({
                                    ...newDevice,
                                    traits: [...newDevice.traits],
                                });
                            });
                            resolve(false);
                        }),
                    detailedOutput: true,
                    allowCustomDevice: false,
                    ...deviceSetup,
                };

                DeviceSetup.setupDevice(device, deviceSetupConfig)
                    .then(() => {
                        // Make sure duplicate device has not been added before callback is run
                        if (
                            !get().devices.some(
                                e => e.serialNumber === device.serialNumber
                            ) &&
                            newDevice.serialport?.path
                        ) {
                            newDevice.correctFirmware = true;
                            newDevice.port = getDevicePort(
                                newDevice.serialport.path
                            );
                            newDevice.parser = newDevice.port.pipe(
                                new Readline({ delimiter: '\n' })
                            );

                            set(state => {
                                state.devices.push({
                                    ...newDevice,
                                    traits: [...newDevice.traits],
                                });
                                state.dirData.deviceDirs.push({
                                    uncalibrated: new Vector3(),
                                    calibrated: new Vector3(),
                                });
                            });

                            persistDevices(
                                get().devices.filter(e => e.isDummy === false)
                            );
                        }
                    })
                    .catch((error: string) => {
                        console.log(error);
                    });
            }

            // Remove from redux store so DeviceSelector component get's cleared
            if (dispatch) {
                dispatch({ type: ActionType.DEVICE_DESELECTED });
            }
        },
        removeDevice: index => {
            if (get().readingState >= ReadingState.CALIBRATION_INITIALIZE) {
                set(state => {
                    state.readingState = ReadingState.OFF;
                });
            }

            set(state => {
                state.devices[index].parser?.unpipe();
                if (state.devices[index].port?.isOpen) {
                    state.devices[index].port?.close();
                }
                state.devices.splice(index, 1);
                state.dirData.deviceDirs.splice(index, 1);
            });
        },
        setCalibrationVectors: (index, vectors) => {
            set(state => {
                state.devices[index].calibrationVectors = vectors;
            });
            persistDevices(get().devices.filter(e => e.isDummy === false));
        },
        setCorrectFirmware: index =>
            set(state => {
                state.devices[index].correctFirmware = true;
            }),
        setDevicePosition: (index, position) => {
            set(state => {
                state.devices[index].position = position;
                state.devices[index].calibrationVectors = [];
            });
            persistDevices(get().devices.filter(e => e.isDummy === false));
        },
        setDeviceRotation: (index, rotation) => {
            set(state => {
                state.devices[index].rotation = rotation;
                state.devices[index].calibrationVectors = [];
            });
            persistDevices(get().devices.filter(e => e.isDummy === false));
        },
        setDeviceDir: (index, dir) =>
            set(state => {
                state.dirData.lastUpdate = index;
                state.dirData.deviceDirs[index] = dir;
            }),
    }))
);
