/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import produce, { Draft } from 'immer';
import { NrfConnectState } from 'pc-nrfconnect-shared';
import create, { State, StateCreator } from 'zustand';

import {
    getPersistedAxesSetting,
    getPersistedCoverageAngle,
    getPersistedCoverageSetting,
    getPersistedDirectionVectorSetting,
    getPersistedLabelSetting,
    persistAxesSetting,
    persistCoverageAngle,
    persistCoverageSetting,
    persistDirectionVectorSetting,
    persistLabelSetting,
} from './persistentStore';

export type ReduxState = NrfConnectState<Record<string, unknown>>;

type SettingsStore = {
    showAxes: boolean;
    showLabels: boolean;
    showCoverage: boolean;
    coverageAngle: number;
    useMovement: boolean;
    showDirectionVectors: boolean;
    useCalibrationVectors: boolean;
    showRawVectors: boolean;
    setUseMovement: (use: boolean) => void;
    toggleAxes: () => void;
    toggleLabels: () => void;
    toggleCoverage: () => void;
    setCoverageAngle: (angle: number) => void;
    toggleDirectionVectors: () => void;
    toggleCalibrationVectors: () => void;
    toggleRawVectors: () => void;
};

const immer = <T extends State>(
    config: StateCreator<T, (fn: (draft: Draft<T>) => void) => void>
): StateCreator<T> => (set, get, api) =>
    config(fn => set(produce(fn) as (state: T) => T), get, api);

export const settingsStore = create<SettingsStore>(
    immer((set, get) => ({
        showAxes: getPersistedAxesSetting(),
        showLabels: getPersistedLabelSetting(),
        showCoverage: getPersistedCoverageSetting(),
        coverageAngle: getPersistedCoverageAngle(),
        showDirectionVectors: getPersistedDirectionVectorSetting(),
        useCalibrationVectors: false,
        showRawVectors: false,
        useMovement: true,
        setUseMovement: use =>
            set(state => {
                state.useMovement = use;
            }),
        toggleAxes: () => {
            set(state => {
                state.showAxes = !state.showAxes;
            });

            persistAxesSetting(get().showAxes);
        },
        toggleLabels: () => {
            set(state => {
                state.showLabels = !state.showLabels;
            });
            persistLabelSetting(get().showLabels);
        },
        toggleCoverage: () => {
            set(state => {
                state.showCoverage = !state.showCoverage;
            });
            persistCoverageSetting(get().showCoverage);
        },
        setCoverageAngle: angle => {
            set(state => {
                state.coverageAngle = angle;
            });
            persistCoverageAngle(get().coverageAngle);
        },
        toggleDirectionVectors: () => {
            set(state => {
                state.showDirectionVectors = !state.showDirectionVectors;
            });
            persistDirectionVectorSetting(get().showDirectionVectors);
        },
        toggleCalibrationVectors: () =>
            set(state => {
                state.useCalibrationVectors = !state.useCalibrationVectors;
            }),
        toggleRawVectors: () =>
            set(state => {
                state.showRawVectors = !state.showRawVectors;
            }),
    }))
);
