/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import { Euler, Vector3 } from 'three';

import { CalibrationVector } from '../stores/deviceTypes';

export function vectorRotationToEuler(rotation: Vector3) {
    return new Euler().setFromVector3(
        rotation.clone().multiplyScalar(Math.PI / 180),
        'YXZ'
    );
}

export function getEulerToGl(rotation: Vector3) {
    const euler = vectorRotationToEuler(rotation);

    return function eulerToGl(vectorLoc: Vector3) {
        return vectorLoc.clone().applyEuler(euler).normalize();
    };
}

export function getEulerToLoc(rotation: Vector3) {
    const euler = new Euler().setFromVector3(
        rotation.clone().multiplyScalar(-Math.PI / 180),
        'ZXY'
    );

    return function eulerToLoc(vectorGl: Vector3) {
        return vectorGl.clone().applyEuler(euler).normalize();
    };
}

export function applyCalibration(
    vector: Vector3,
    calibrationVectors: CalibrationVector[]
) {
    const angles: number[] = [];
    const lerpOffset = new Vector3();

    if (calibrationVectors.length === 0) return vector.clone();

    calibrationVectors.forEach(({ measured }) => {
        const angle = measured.angleTo(vector);
        angles.push(angle);
    });
    const avgAngle =
        angles.reduce((tot, curr) => tot + curr, 0) / angles.length;
    for (let i = 0; i < calibrationVectors.length; i += 1) {
        const scalar = 1 - angles[i] / avgAngle;
        if (scalar > 0) {
            lerpOffset.add(
                calibrationVectors[i].offset.clone().multiplyScalar(scalar)
            );
        }
    }
    return vector.clone().add(lerpOffset).normalize();
}

function averagePoints(points: Vector3[]) {
    const avgPoint = points[0].clone();
    for (let i = 1; i < points.length; i += 1) {
        avgPoint.add(points[i]);
    }
    avgPoint.divideScalar(points.length);
    return avgPoint;
}

export function getClosestPointsVectors(
    positions: [Vector3, Vector3],
    unitVectors: [Vector3, Vector3]
): [Vector3, Vector3] {
    const n = unitVectors[0].clone().cross(unitVectors[1]);
    const n2 = unitVectors[1].clone().cross(n);
    const c1 = positions[0]
        .clone()
        .add(
            unitVectors[0]
                .clone()
                .multiplyScalar(
                    positions[1].clone().sub(positions[0]).dot(n2) /
                        unitVectors[0].clone().dot(n2)
                )
        );

    const n1 = unitVectors[0].clone().cross(n);
    const c2 = positions[1]
        .clone()
        .add(
            unitVectors[1]
                .clone()
                .multiplyScalar(
                    positions[0].clone().sub(positions[1]).dot(n1) /
                        unitVectors[1].clone().dot(n1)
                )
        );

    return [c1, c2];
}

export function getPosition(positions: Vector3[], unitVectors: Vector3[]) {
    const pointArr: Vector3[] = [];
    for (let i = 0; i < positions.length; i += 1) {
        for (let j = 0; j < positions.length; j += 1) {
            if (i < j) {
                const points = getClosestPointsVectors(
                    [positions[i], positions[j]],
                    [unitVectors[i], unitVectors[j]]
                );

                // If lines are parallell
                // It's sufficient to check a single component of either vector as they will all be NaN
                if (!Number.isNaN(points[0].x)) {
                    pointArr.push(averagePoints(points));
                }
            }
        }
    }
    if (pointArr.length === 0) {
        return new Vector3();
    }

    return averagePoints(pointArr);
}
