/* Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import SerialPort from 'serialport';
import { Vector3 } from 'three';

export const { Readline } = SerialPort.parsers;

export function getDevicePort(serialPath: string) {
    return new SerialPort(serialPath, {
        baudRate: 115200,
        parity: 'none',
        stopBits: 1,
        dataBits: 8,
    });
}

export interface DataFrame {
    azimuth: number;
    elevation: number;
}

function calculateArrowPosition({ azimuth, elevation }: DataFrame) {
    elevation = 90 - elevation;
    azimuth = 360 - azimuth;

    const tan = Math.tan((elevation * Math.PI) / 180);
    const rad = (azimuth * Math.PI) / 180;
    const x = Math.cos(rad) / tan;
    const y = Math.sin(rad) / tan;
    const z = 1;
    return new Vector3().fromArray([x, y, z]).normalize();
}

export function parseFrame() {
    let currentFrame: DataFrame;
    return function parseMessage(message: string) {
        // Messages can have a variable number of carriage returns at the end,
        // hence we can't compare for string equality.
        if (message.startsWith('DF_BEGIN')) {
            currentFrame = { azimuth: 0, elevation: 0 };
            return;
        }

        // We can't just start reading at any point, since the device
        // could be in the middle of a frame. Hence we exit early if
        // the frame hasn't been initialised.
        if (!currentFrame) return;

        if (message.startsWith('DF_END')) {
            currentFrame.azimuth = 360 - currentFrame.azimuth;

            return calculateArrowPosition(currentFrame);
        }

        const [head, data] = message.split(':');

        // Irrelevant frame data
        if (head === 'IQ') return;

        const value = parseInt(data, 10);

        switch (head) {
            // Filtered values
            case 'KE': {
                currentFrame.elevation = value;
                return;
            }
            case 'KA': {
                currentFrame.azimuth = value;
            }
            // Irrelevant frame data
            // case 'ME':
            // case 'MA':
            // case 'FR':
            // case 'SW':
            // case 'RR':
            // case 'SS':
        }
    };
}
